#include "barrelfish/barrelfish.h"
#include "barrelfish_kpi/paging_target.h"
#include "barrelfish/except.h"

#include <stdio.h>

#define EXCEPTION_STACK_SIZE        (10 * X86_32_BASE_PAGE_SIZE)

// this will be the base address of our test array. The value is arbitrary. But
// is safe (hope so..), i.e, won't be touched by the kernel or other applications
#define SAFE_ADDR                   (0x60000000)
#define SAFE_ADDR2                  (0x68000000)

#define TEST_SIZE                   (2 * X86_32_BASE_PAGE_SIZE)

static struct pmap* pmap;

// test array length
const size_t len = TEST_SIZE / sizeof(int);

// the element number that a page contain
const size_t page_len = X86_32_BASE_PAGE_SIZE / sizeof(int);

    void
exception_handler(enum exception_type type, int subtype, void* addr,
        arch_registers_state_t* regs, arch_registers_fpu_state_t* fpu) 
{
    // we only handle page fault
    if (type != EXCEPT_PAGEFAULT) {
        return;
    }

    lvaddr_t page_base = (lvaddr_t)ROUND_DOWN((uintptr_t)addr, X86_32_BASE_PAGE_SIZE);
    errval_t err;

    switch(subtype) {
        // we only handle write fault
        case PAGEFLT_WRITE:
            // map the fault page as RW
            err = pmap->f.modify_flags(pmap, vspace_lvaddr_to_genvaddr(page_base), 
                    X86_32_BASE_PAGE_SIZE, VREGION_FLAGS_READ_WRITE, NULL);
            if (err_is_fail(err)) {
                debug_printf("modify_flags failed: %s\n", err_getstring(err));
            }
            break;

        default:
            debug_printf("other page fault type.\n");
            break;
    }
}

int
monolithic_frame(void) {

    debug_printf("begin monothelic_frame test...\n");

    struct capref frame;

    // first test, allocate the array as a whole frame
    errval_t err = frame_alloc(&frame, TEST_SIZE, NULL);
    if (err_is_fail(err)) {
        debug_printf("frame_alloc failed: %s\n", err_getstring(err));
        return EXIT_FAILURE;
    }

    err = pmap->f.map(pmap, vspace_lvaddr_to_genvaddr((lvaddr_t)SAFE_ADDR), 
            frame, 0, TEST_SIZE, VREGION_FLAGS_READ, NULL, NULL);
    if (err_is_fail(err)) {
        debug_printf("map failed: %s\n", err_getstring(err));
        return EXIT_FAILURE;
    }

    int* a = (int*)SAFE_ADDR;

    for (int i = 0; i < len; i++) {
        a[i] = i;
    }

    debug_printf("array init done!\n");

    for (int i = 0; i < len; i++) {
        if (a[i] != i) {
            debug_printf("array check fail, a[%d] = %d\n", i, a[i]);

            // we should have a[i] == i, but it appears not. So here I just print the values
            // of some element
            int j;
            for (j = 0; j < page_len; j++) {
                if (a[j] != a[j+page_len]) {
                    break;
                }
            }

            if (j == page_len) {
                debug_printf("two page's content are identical\n");

                for (int k = 0; k < 4; k++) {
                    printf("a[%d] = %d, a[%d] = %d\n", k, a[k], k+page_len, a[k+page_len]);
                }
            }

            debug_printf("checking if the two virtual page share the same physical page... ");
            a[0] = 0xdeadbeef;
            printf("%s\n", a[page_len] == 0xdeadbeef? "yes": "no");
            break;
        }
    }

    return EXIT_SUCCESS;
}

int
one_page_one_frame(void) {
    debug_printf("begin one_page_one_frame test...\n");

    struct capref frame;
    errval_t err;

    // second test, allocate one page once
    int frame_num = TEST_SIZE / X86_32_BASE_PAGE_SIZE;
    for (int i = 0; i < frame_num; i++) {
        err = frame_alloc(&frame, X86_32_BASE_PAGE_SIZE, NULL);
        if (err_is_fail(err)) {
            debug_printf("frame_alloc failed: %s\n", err_getstring(err));
            return EXIT_FAILURE;
        }

        err = pmap->f.map(pmap, vspace_lvaddr_to_genvaddr((lvaddr_t)SAFE_ADDR2+i*X86_32_BASE_PAGE_SIZE), 
                frame, 0, X86_32_BASE_PAGE_SIZE, VREGION_FLAGS_READ, NULL, NULL);
        if (err_is_fail(err)) {
            debug_printf("map failed: %s\n", err_getstring(err));
            return EXIT_FAILURE;
        }
    }

    int* a = (int*)SAFE_ADDR2;

    for (int i = 0; i < len; i++) {
        a[i] = i;
    }

    debug_printf("array init done\n");

    for (int i = 0; i < len; i++) {
        assert(a[i] == i);
    }

    debug_printf("array check done!\n");

    return EXIT_SUCCESS;
}

int main(int argc, char* argv[]) {
    (void)argc;
    (void)argv;

    // allocate a exception stack.
    void* exception_stack_base = malloc(EXCEPTION_STACK_SIZE);
    void* exception_stack_top = exception_stack_base + EXCEPTION_STACK_SIZE;

    thread_set_exception_handler(exception_handler, NULL, exception_stack_base, 
            exception_stack_top, NULL, NULL);

    pmap = vspace_get_pmap(get_current_vspace());

    if (monolithic_frame() != EXIT_SUCCESS) {
        return EXIT_FAILURE;
    }

    if (one_page_one_frame() != EXIT_SUCCESS) {
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
